%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for 68020 processor
% Author:       H. Melenk
% Created:      25 January 1989
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the 68020 processor.
% These are the open codes which embed the extended arithmetic
% operations such that they are accessible from compiled LISP top level.
% They map the 2*30 bit operations needed by BIG to the 2*32 bit operations
% of the 68020 by shifting etc.
(fluid '(*second-value*))


(put 'wtimesdouble 'opencode
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The high order bits are placed in fluid variable.
     '((MULSLONG (reg 2)(reg 1)(reg 3))
      % now we have 32 low bits in REG1, 32 high bits in REG3
       (*WSHIFT (reg 3) 2)
       (*MOVE (reg 1) (reg 2))
       (*WSHIFT (reg 2) -30)
       (*WOR (reg 3)(reg 2))
       (*MOVE (reg 3)($FLUID *second-value*))
       (*WSHIFT (reg 1) 2)
       (*WSHIFT (reg 1) -2) ))

(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '((MULSLONG (reg 2)(reg 1)(reg 3))
      (*MOVE (reg 3)($FLUID *second-value*))
      ))

(put 'wquotientdouble 'opencode
      % called with a double length number in params 1 and 2
      % and a single length number in par 3.
      % Result is the single length quotient.
      % the remainder is placed in a fluid variable.
      '(  % adjusting the words first
        (*MOVE (reg 1) (reg 4))
        (*WSHIFT (reg 1) -2)
        (*WSHIFT (reg 4) 30)
        (*WOR (reg 2) (reg 4))
          % now we can divide and spread the results
        (DIVSLONG (reg 3) (reg 2) (reg 1))
        (*MOVE (reg 1) ($FLUID *second-value*))
        (*MOVE (reg 2) (reg 1))))

(put 'addcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WPLUS2 (reg 1)(reg 2))
         (*MOVE (reg 1)(reg 3))
         (*WSHIFT (reg 3) -30)
         (*MOVE (reg 3) ($FLUID carry*))
         (*WSHIFT (reg 1) 2)
         (*WSHIFT (reg 1) -2)))

(put 'subcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         (*MOVE (reg 1) (reg 2))
         (*WSHIFT (reg 2) -31)
         (*MOVE (reg 2) ($FLUID carry*))
         (*WSHIFT (reg 2) 30)
         (*WPLUS2 (reg 1) (reg 2))
       ))


(put 'subcarry 'opencode  % fast variant for 68020
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         (*MOVE (reg 1) (reg 2))
         (ROL!.l 1 (reg 2))
         (*MOVE 1 (reg 3))
         (*WAND (reg 2)(reg 3))
         (*MOVE (reg 2) ($FLUID carry*))
         (ROR!.l 2 (reg 2))
         (*WPLUS2 (reg 1) (reg 2))
       ))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   functions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.

%   get the value for bbase* into a register
(put 'bbase** 'opencode
      '((*MOVE 1 (reg 1))
        (*WSHIFT (reg 1) 30)))

(put 'bbase** 'opencode % fast variant for 68020
      '((*MOVE 1 (reg 1))
        (ROR!.l 2 (reg 1))))

%   get the value for logicalbits* into a register
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
        (*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*WSHIFT (reg 1) 2)
        (*WSHIFT (reg 1) -2)))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -30)))

(put 'bbase** 'destroys '((reg 1)))
(put 'logicalbits** 'destroys '((reg 1)))
(put 'remainder-bbase 'destroys '((reg 1)))
(put 'quotient-bbase 'destroys '((reg 1)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   68020 instructions supporting extended mult. and div.
%

%   Instruction for multiplying 32*32=64 bits (68020)
%   (MULSLONG a b c)
%      only registers allowed for a,b and c
%      executes b := a*b (bits 0-31)
%               c := a*b (bits 32-63)
(put 'MULSLONG 'Size 1)
(put 'MULSLONG 'INSTRUCTIONDEPOSITFUNCTION 'deposit64oplong)
(put 'MULSLONG 'INSTRUCTIONLENGTH 4)
(put 'MULSLONG 'opcodeValue 2#0100110000)

%   Instruction for multiplying 32*32=32 bits (68020)
%   (MULS!.L a b c)
%      only registers allowed for a,b and c
%      executes b := a*b (bits 0-31)
(put 'MULS!.L 'Size 1)
(put 'MULS!.L 'INSTRUCTIONDEPOSITFUNCTION 'deposit64opl)
(put 'MULS!.L 'INSTRUCTIONLENGTH 4)
(put 'MULS!.L 'opcodeValue 2#0100110000)

%   Instruction for dividing a 64 bits by a 32 bits (68020)
%   (DIVSLONG a b c)
%      only registers allowed for a,b and c
%      executes b := (b,c) / a
%               c := remainder ( (b,c) , a)
(put 'DIVSLONG 'Size 1)
(put 'DIVSLONG 'INSTRUCTIONDEPOSITFUNCTION 'deposit64oplong)
(put 'DIVSLONG 'INSTRUCTIONLENGTH 4)
(put 'DIVSLONG 'opcodeValue 2#0100110001)

%   Instruction for dividing a 32 bits by a 32 bits (68020)
%   (DIVSL.L a b c)
%      only registers allowed for a,b and c
%      executes c := c / a
%               b := remainder ( c , a)
(put 'DIVSL!.L 'Size 1)
(put 'DIVSL!.L 'INSTRUCTIONDEPOSITFUNCTION 'deposit64opdivl)
(put 'DIVSL!.L 'INSTRUCTIONLENGTH 4)
(put 'DIVSL!.L 'opcodeValue 2#0100110001)



(de deposit64oplong(instr)
(prog (SourceMode SourceRegisterNumber DlRegisterNumber DhRegisterNumber
       DestMode op)
    (setq op (get (first Instr) 'opcodeValue))
    (setq SourceMode (RegOperand       (second Instr)))
    (setq SourceRegisterNumber OperandRegisterNumber*)

    (setq DestMode (RegOperand       (third Instr)))
    (setq DlRegisterNumber OperandRegisterNumber*)

    (setq DestMode (RegOperand       (third (cdr Instr))))
    (setq DhRegisterNumber OperandRegisterNumber*)

    (setq InstructionSize* 4)

    (DepositHalfWord (plus
                     (LSH op 6)
                     (LSH SourceMode 3)
                     SourceRegisterNumber))

    (Deposithalfword (plus
                     (LSH DlRegisterNumber 12)
                     (LSH 2#110000000 3)
                     DhRegisterNumber))))


(de deposit64opl(instr)
(prog (SourceMode SourceRegisterNumber DlRegisterNumber DhRegisterNumber
       DestMode op)
    (setq op (get (first Instr) 'opcodeValue))
    (setq SourceMode (RegOperand       (second Instr)))
    (setq SourceRegisterNumber OperandRegisterNumber*)

    (setq DestMode (RegOperand       (third Instr)))
    (setq DlRegisterNumber OperandRegisterNumber*)

    %(setq DestMode (RegOperand       (third (cdr Instr))))
    %(setq DhRegisterNumber OperandRegisterNumber*)

    (setq InstructionSize* 4)

    (DepositHalfWord (plus
                     (LSH op 6)
                     (LSH SourceMode 3)
                     SourceRegisterNumber))

    (Deposithalfword (plus
                     (LSH DlRegisterNumber 12)
                     (LSH 2#100000000 3)
                     ))))%DhRegisterNumber))))


(de deposit64opdivl(instr)
(prog (SourceMode SourceRegisterNumber DlRegisterNumber DhRegisterNumber
       DestMode op)
    (setq op (get (first Instr) 'opcodeValue))
    (setq SourceMode (RegOperand       (second Instr)))
    (setq SourceRegisterNumber OperandRegisterNumber*)

    (setq DestMode (RegOperand       (third  Instr)))
    (setq DhRegisterNumber OperandRegisterNumber*)

    (setq DestMode (RegOperand       (third (cdr Instr))))
    (setq DlRegisterNumber OperandRegisterNumber*)

    (setq InstructionSize* 4)

    (DepositHalfWord (plus
                     (LSH op 6)
                     (LSH SourceMode 3)
                     SourceRegisterNumber))

    (Deposithalfword (plus
                     (LSH DlRegisterNumber 12)
                     (LSH 2#100000000 3)
                     DhRegisterNumber))))



